// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

#include "utility.h"
#include <string>
#include <cmath>
#include<vector>
#include<numeric>
#include <iostream>
#include <sstream>
#include <iomanip>
#include "Dal.h"

#include "SasDalUtils.h"

//#include "dal_utilities.h"

#include "DataSubspace.h"

#include"f90cxx.h"




/**************************************************************
*
*
***************************************************************/

bool isNull(int8 value)
{  

    if( value == 127)
      return true;
    else
    return false;    

}
bool isNull(int16 value)
{  

    if( value == 32767)
      return true;
    else
    return false;    

}

bool isNull(int32 value)
{  

       
    stringstream ss;
    ss << value;
    string text = ss.str();
 
        
    if ( (int)text.find("nan") >= 0) return true;
    if ( (int)text.find("NAN") >= 0) return true;
    if ( (int)text.find("null") >= 0) return true;
    if ( (int)text.find("NULL") >= 0) return true;

    return false;    


}




bool isNull( float value)
{  



    stringstream ss;
    ss << value;
    string text = ss.str();
 


        
    if ( (int)text.find("nan") >= 0) return true;
    if ( (int)text.find("NAN") >= 0) return true;
    if ( (int)text.find("null") >= 0) return true;
    if ( (int)text.find("NULL") >= 0) return true;

    return false;    

};

bool isNull( double value)
{  

   
    stringstream ss;
    ss << value;
    string text = ss.str();
    
    if ( (int)text.find("nan") >= 0) return true;
    if ( (int)text.find("NAN") >= 0) return true;
    if ( (int)text.find("null") >= 0) return true;
    if ( (int)text.find("NULL") >= 0) return true;

    return false;    

};

bool isNull( char * value)
{  

    stringstream ss;
    ss << value;
    string text = ss.str();
         
    if ( (int)text.find("nan") >= 0) return true;
    if ( (int)text.find("NAN") >= 0) return true;
    if ( (int)text.find("null") >= 0) return true;
    if ( (int)text.find("NULL") >= 0) return true;

    return false;    

};


/*******************************************************************
*
*
*
*******************************************************************/
double distance(double r1, double d1, double r2, double d2)
{



    return acos( sin(d1) * sin(d2) + cos(d1) * cos(d2) 
	            * cos(r1 - r2) ) * radiansToArcSecs; 


};
/*************************************************************************
*
*
*
***************************************************************************/
double Error(double dx, double dy)
{
  if (dx > dy)
    return dx * sqrt(1.0 + (dy / dx) * (dy / dx));
  else
    return dy * sqrt(1.0 + (dx / dy) * (dx / dy));
} 





/***************************************************************
*
*
*
*
****************************************************************/
float Mean(float *data, long nData)
{

    double mean = 0.0;
    if (nData == 0) return 0.0;

    for( long i = 0; i < nData; i++)
        mean += (double)data[i];

    return (float)(mean / (double)nData);

}



/***************************************************************
*
*
****************************************************************/
float Median(float *data, long nData)
	{
    float median;


    if (nData == 0) return 0.0;
    if (nData == 1) return data[0]; 



  
    int n2 = nData / 2;

    if ( n2 * 2 == nData ) // Even number of points
        median = (data[n2 - 1] + data[n2]) / 2.0;
    else
      median = data[n2];

    return median;
}    

/***************************************************************
*
*
****************************************************************/
//float Percentile(vector<float> &data, float perCent)
//{

//   long size = data.size();
    
//    long index = (long)(perCent * (float)size + 0.00001) - 1;
//    index = max((long)0, index);
//    index = min(index, size - 1); 
     
//    cout << "n=" << size << " percent=" << perCent << " index=" << index << " percentile=" << data[index] << endl;
//    cout << "data[0]=" << data[0] << " data[" << (size - 1) << "]" << data[size - 1] << endl;
//    return data[index];


//}
/***************************************************************
*
*
****************************************************************/
float Percentile(float *data, long size, float perCent)
{

 
    
    long index = (long)(perCent * (float)size + 0.00001) - 1;
    index = max((long)0, index);
    index = min(index, size - 1); 

    return data[index];

}


/*******************************************************
*
*
*******************************************************/


vector<vector<int8> >  loadInt8Image(const DataSet *set, const char * name, int &nx, int &ny)
{


    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
    const Array* array = set->array(name);
 

    const ArrayData * arraydata = array->data();
    const int8 * theData = arraydata->int8Data();
 

    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
 
    longVec size = array->size();
   



    nx = (int)size[1];
    ny = (int)size[0];


    vector< vector<int8> > image(nx); 


    for (int i = 0; i < nx; i++)
    {
        image[i].resize(ny); 
        for (int j = 0; j < ny; j++)
        {
	  image[i][j] = theData[i + (j * nx)];  
	 

        }

  
    }


    delete arraydata;
   

    return image;
}

/*******************************************************
*
*
*
*******************************************************/


vector<vector<int16> >  loadInt16Image(const DataSet *set, const char * name, int &nx, int &ny)
{


    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
    const Array* array = set->array(name);
 

    const ArrayData * arraydata = array->data();
    const int16 * theData = arraydata->int16Data();
 

    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
 
    longVec size = array->size();
   



    nx = (int)size[1];
    ny = (int)size[0];


    vector< vector<int16> > image(nx); 


    for (int i = 0; i < nx; i++)
    {
        image[i].resize(ny); 
        for (int j = 0; j < ny; j++)
        {
	  image[i][j] = theData[i + (j * nx)];  
	 

        }

  
    }


    delete arraydata;
   

    return image;
}





/*******************************************************
*
*
*
*******************************************************/


vector<vector<float> >  loadFloatImage(const DataSet *set, const char * name, int &nx, int &ny)
{


    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
    const Array* array = set->array(name);
 

    const ArrayData * arraydata = array->data();
    const float * theData = arraydata->real32Data();
 

    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
 
    longVec size = array->size();
   



    nx = (int)size[1];
    ny = (int)size[0];


    vector< vector<float> > image(nx); 


    for (int i = 0; i < nx; i++)
    {
        image[i].resize(ny); 
        for (int j = 0; j < ny; j++)
        {
	  image[i][j] = theData[i + (j * nx)];  
	 

        }

  
    }


    delete arraydata;
   

    return image;
}



/********************************************************************************
*
*
********************************************************************************/

bool modifyInt16Image(DataSet *set, const char * name, const vector<vector<int16> > &image)
{

    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
    Array *array = set->array(name);
 

    ArrayData * arraydata = array->data();
    int16 * theData = arraydata->int16Data();
 

    /*****************************************************************
    * Get the size of the image
    *****************************************************************/

    int nx = (int)image.size();
    int ny = (int)image[0].size();


 


    for (int i = 0; i < nx; i++)
    {
    
        for (int j = 0; j < ny; j++)
        {
	   theData[i + (j * nx)] = image[i][j];  
	 

        }

  
    }


    delete arraydata;
   

    return true;
}



/*******************************************************
*
*
*******************************************************/


vector<vector<int8> >  loadInt8Image(const string inSet, const char * name, int &nx, int &ny)
{


    DataSet *set = dataSetServer->open(inSet, Dal::Read);
 


    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
    const Array* array = set->array(name);
 

    const ArrayData * arraydata = array->data();
    const int8 * theData = arraydata->int8Data();
 

    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
 
    longVec size = array->size();
   



    nx = (int)size[1];
    ny = (int)size[0];


    vector< vector<int8> > image(nx); 


    for (int i = 0; i < nx; i++)
    {
        image[i].resize(ny); 
        for (int j = 0; j < ny; j++)
        {
	  image[i][j] = theData[i + (j * nx)];  
	 

        }

  
    }


    delete arraydata;
   
    dataSetServer->close( set ); // Finished with the file

    return image;
}




/*******************************************************
*
*
*******************************************************/


vector<vector<int16> >  loadInt16Image(const string inSet, const char * name, int &nx, int &ny)
{


    DataSet *set = dataSetServer->open(inSet, Dal::Read);
 


    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
    const Array* array = set->array(name);
 

    const ArrayData * arraydata = array->data();
    const int16 * theData = arraydata->int16Data();
 

    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
 
    longVec size = array->size();
   



    nx = (int)size[1];
    ny = (int)size[0];


    vector< vector<int16> > image(nx); 


    for (int i = 0; i < nx; i++)
    {
        image[i].resize(ny); 
        for (int j = 0; j < ny; j++)
        {
	  image[i][j] = theData[i + (j * nx)];  
	 

        }

  
    }


    delete arraydata;
   
    dataSetServer->close( set ); // Finished with the file

    return image;
}



/*******************************************************
*
*
*******************************************************/


vector<vector<float> >  loadFloatImage(const string inSet, const char * name, int &nx, int &ny)
{


    DataSet *set = dataSetServer->open(inSet, Dal::Read);
 


    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
    const Array* array = set->array(name);
 

    const ArrayData * arraydata = array->data();
    const float * theData = arraydata->real32Data();
 

    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
 
    longVec size = array->size();
   



    nx = (int)size[1];
    ny = (int)size[0];


    vector< vector<float> > image(nx, vector<float>(ny)); 


    for (int i = 0; i < nx; i++)
    {
      
        for (int j = 0; j < ny; j++)
        {
            

	  image[i][j] = theData[i + (j * nx)];  
	  //  cout << image[i][j] << endl; 	 
	
        }

  
    }


    delete arraydata;
   
    dataSetServer->close( set ); // Finished with the file

    return image;
}



/**************************************************************************
*
* Adds the image (float) to the data set with the block name name
*
************************************************************************/
void addBlock(DataSet *set, string name, vector< vector<float> > &image, int position)
{


    int nx = (int)image.size();
    int ny = (int)image[0].size();
    longVec size(2);
    size[0] = ny;
    size[1] = nx;




    Array * array  = set->addArray(name,
				   Dal::Real32,
				   size, "","", position);
 


    //		block = set -> addArray(
    //					name.str(),
    //					Array::Int32,
    //					vector<unsigned long>(size,size+3),
    //					label.str(),
    //					units.str(),
    //					blockDesc[ i ].blockNumber );
		




    ArrayData * arraydata = array->data();
    real32 * thedata = arraydata->real32Data();
 

 
 


    

    for (int i = 0; i < nx; i++)
    {
        for (int j = 0; j < ny; j++)
	{
          if (image[i][j] >= 0.0)
	      thedata[i + (j * nx)] = image[i][j];
          else 
	  {
	      arraydata->setToNull( i + (j * nx) );
	    
          }
	}
    }

    delete arraydata;

}





/**************************************************************************
*
* Adds the image (float) to the data set with the block name name
*
************************************************************************/
void addBlock(DataSet *set, string name, vector< vector<double> > &image, int position)
{


    int nx = (int)image.size();
    int ny = (int)image[0].size();
    longVec size(2);
    size[0] = ny;
    size[1] = nx;




    Array * array  = set->addArray(name,
				   Dal::Real64,
				   size, "", "", position);
    ArrayData * arraydata = array->data();
    real64 * thedata = arraydata->real64Data();
 
    for (int i = 0; i < nx; i++)
    {
        for (int j = 0; j < ny; j++)
	{
          if (image[i][j] >= 0.0)
	      thedata[i + (j * nx)] = image[i][j];
          else 
	  {
	      arraydata->setToNull( i + (j * nx) );
	    
          }
	}
    }

    delete arraydata;

}

/**************************************************************************
*
* Adds the image (integer 16) to the data set with the block name name
*
************************************************************************/
void addBlock(DataSet *set, string name, vector< vector<int16> > &image, int position)
{


    int nx = (int)image.size();
    int ny = (int)image[0].size();
    longVec size(2);

    size[0] = ny;
    size[1] = nx;


    Array * array  = set->addArray(name,
				   Dal::Int16,
				   size, "", "", position);
    ArrayData * arraydata = array->data();
    int16 * thedata = arraydata->int16Data();
 
    array->nullValue(-1);
     

    for (int i = 0; i < nx; i++)
    {
        for (int j = 0; j < ny; j++)
	{
	      if (image[i][j] >= 0)
	      {
	          thedata[i + (j * nx)] = image[i][j];
              }
	      else 
	      {
	          arraydata->setToNull( i + (j * nx) );
	    
	      }
	}
    }

    delete arraydata;

}

/**********************************************************
*
*
**********************************************************/
void normalizeImage(vector<vector<float> > & image1, vector<vector<float> > &image2)
{


    for(unsigned i = 0; i < image1.size(); i++)
    {
        for(unsigned j = 0; j < image1[i].size(); j++)
        {
	    if ( ( image1[i][j] > 0.0 && !isNull(image1[i][j] ) ) &&
                 ( image2[i][j] > 0.0 && !isNull(image2[i][j] ) ) )
	    {
	        image1[i][j] /= image2[i][j];
            }
        }
    } 

 
}






/************************************************************************
*
*
*
*************************************************************************/


F90MODPROCPROTO( void, sla, sla_cldj)(int *, int *, int *, double *, int *);



/*******************************************************
* ModifiedJulianDate - Returns the modified Julian date
* from the string date representation
********************************************************/
double ModifiedJulianDate(const string dateString)
{

    int year = 0;           // Stores the year from the date string
    int month = 0;          // Stores the month from the date string
    int day = 0;            // Stores the day from the date string  
    int hour = 0;           // Stores the hour from the date string
    int minute = 0;         // Stores the minute from the date string  
    int second = 0;         // Stores the second from the date string
    char c;                 // Used for date and time extraction 

 
    double JulianDay = 0;    
    int flag = 0;


    istringstream ss(dateString);
    ss >> year >> c >> month >> c >> day >> c >> hour >> c >> minute >> c >> second;



    //    sscanf(dateString, "%4d, %1c, %2d, %1c, %2d, %1c, %2d, %1c, %2d, %1c, %2d",
    //     &year, &c, &month, &c, &day, &c, &hour, 
    //     &c, &minute, &c, &second);     

    // Get the modified Julian date for this date
 //   call sla_CLDJ(year, month, day, JulianDay, flag)

    F90MODPROC( sla, sla_cldj)(&year, &month, &day, &JulianDay, &flag);



    //  Need to add on the fraction of a day corresponding to the time
    JulianDay  += (hour + 
         (double(minute) + double(second) / 60.0) / 60.0) / 24.0;



    //   cout << "OMWCS.cc ModifiedJulianDate line 956- dateString=" << dateString << endl;
    // cout << "Year=" << year << " month=" << month << " hour=" << hour << " minute=" << minute << " second=" << second << endl;
    // cout << "Julian day=" << setprecision(20) << JulianDay << endl; 

    return JulianDay;

}











