// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#include "Dal.h"
#include <iostream>
#include "SasDalUtils.h"
#include "dal_utilities.h"
#include "DataSubspace.h"
#include<cstdlib>
#include<vector>
#include<cmath>
#include "createimage.h"

#include <f90cxx.h>



/******************************************************
*
*
*
*
*
*******************************************************/
bool LoadVectorFloatColumnData(const Table *table, 
                      const string& columnName, 
		      int &size, vector<float> &data);

bool LoadVectorDoubleColumnData(const Table *table, 
                      const string& columnName, 
		      int &size, vector<double> &data);
void loadData(string file, vector<double> &ra, vector<double> &dec);


bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
	             vector<real32> &data);

bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
	             vector<real64> &data);

bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
	             vector<int8> &data);

bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
	             vector<int16> &data);


bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
	             vector<int32> &data);


void setColumnData( const int size, const ColumnData *columnData, const vector<int8>& data);
void setColumnData( const int size, const ColumnData *columnData, const vector<int16>& data);
void setColumnData( const int size, const ColumnData *columnData, const vector<int32>& data);
void setColumnData( const int size, const ColumnData *columnData, const vector<real32>& data);
void setColumnData( const int size, const ColumnData *columnData, const vector<real64>& data);

/**********************************************************************
*
*
* First test is to create a list of sources and then add the quality flags 
* to the sky image
*
*
**********************************************************************/ 

void addSkyattributes(string file, int x0, int y0, float ra, float dec);

int main()
{


    /*****************************************
    *
    * Initiate the random-number generator
    *
    ******************************************/
    srand ( time(NULL) );
 
    /***********************************************
    *
    * Random image dimensions
    *
    ***********************************************/
    int nx = max(100, rand() % 500 + 1);
    int ny = max(100, rand() % 500 + 1);

    nx = ny = 100;

    /***********************************************
    *
    * Random background
    *
    ***********************************************/
    float background = 9.0 + rand() % 100;

    cout << "nx=" << nx << " ny=" << ny << " background=" << background << endl;




    /*******************************************************************
    *
    * Create the 2-d image and set all pixels equal to the background
    *
    *******************************************************************/
    vector< vector<float> > image(nx, vector<float>(ny, background) );



    /****************************************************
    *
    * Add a random number of sources at rando positions
    *
    *****************************************************/
    int noSources = 10 + rand() % 10;


    for (int i = 0; i < noSources; i++)
    {
        float xPos = float(rand() % nx + 1);
        float yPos = float(rand() % ny + 1);
        float height = 10.0 + float(rand() % 1000);

        addSource(image, xPos, yPos, height);

    }

    int x0 = nx / 2;
    int y0 = ny / 2;


    /****************************************************************
    *
    * Create the image file
    *
    ****************************************************************/
    createImageFile("testimage.fits", image, 0, 0);


    /***********************************************
    *
    * Random OM pointing direction
    *
    ***********************************************/
    float ra = rand() % 300;
    float dec = -50.0 + rand() % 100;
    ra = 45.0;
    dec = 45.0;
    const int noFiles = 2; // The number of image files to create

    for(int i = 1; i <= noFiles; i++)
    {
        char file[100];
        sprintf(file, "testimage%02d.fits", i);
        int k = 5 * (i - 1);
        createImageFile(file, image, k, -k);
       
        /****************************************************************
        *
        * Create sky images
        *
        ****************************************************************/
        sprintf(file, "skyimage%02d.fits", i);
   
        dataSetServer->copy("testimage.fits", file);
        
        addSkyattributes(file, x0, y0, ra, dec);

	x0 += 2;  // Add small image displacement
	y0 += 2;
  
    }

    return 0;

}
/***********************************************************
*
*
*
*
**********************************************************/
void loadData(string file, vector<double> &ra, vector<double> &dec)
{

    DataSet *dataSet = dataSetServer->open( file, Dal::Read);

    if (!dataSet)
    {


    }    
 
    const Table *table = dataSet->table( "SRCLIST", Dal::Read);
    if( !table) {};

    /**********************************************************************
    *
    *
    **********************************************************************/
  
    int size = 0;

 
    if ( !LoadVectorDoubleColumnData( table, "RA_CORR", size, ra) )
    {

    

    }

 
    if ( !LoadVectorDoubleColumnData( table, "DEC_CORR", size, dec) )
    {

    

    }

}
/********************************************************
*
*
******************************************************/
bool LoadVectorFloatColumnData(const Table *table, 
                      const string& columnName, 
                      int &size, vector<float> &data)
{
     
    size = table->rows();
   
    if (table->hasColumn(columnName))
    { 

        const ColumnData *columnData = table->column(columnName)->data();
        if (!columnData) exit(0);
	

      
        float *xData = columnData->real32Data();
       


	for (int i=0; i< size; i++) 
        {
	  data.push_back( xData[i] ); 
	 
	}
     }
         
   
    return true;

};

/********************************************************
*
*
******************************************************/
bool LoadVectorDoubleColumnData(const Table *table, 
                      const string& columnName, 
                      int &size, vector<double> &data)
{
       
    size = table->rows();
   
    if (table->hasColumn(columnName))
    { 
        const ColumnData *columnData = table->column(columnName)->data();
        if (!columnData) exit(0);
	

      
        double *xData = columnData->real64Data();
       


	for (int i=0; i< size; i++) 
        {
	  data.push_back( xData[i] ); 
	 
	}
     }
         
   
    return true;
};



/********************************************************
*
*
******************************************************/
bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
                     vector<real32> &data)
{
     

  
    // Get the size of the table
    vector<unsigned long> size(1);
    size[0] = 1;
    size[1] = nData;
  
   
    // Create the column
    Column *column = table->addColumn( columnName, Dal::Real32, comment, units , size);
 
    ColumnData *columnData = column->data();

    setColumnData( ( int)size[1], columnData, data);


    delete columnData;

          
    return true;

};



/********************************************************
*
*
******************************************************/
bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
                     vector<real64> &data)
{
     

  
    // Get the size of the table
    vector<unsigned long> size(1);
    size[0] = 1;
    size[1] = nData;
  
   
    // Create the column
    Column *column = table->addColumn( columnName, Dal::Real64, comment, units , size);
 
    ColumnData *columnData = column->data();

    setColumnData( ( int)size[1], columnData, data);


    delete columnData;


          
    return true;
};

/********************************************************
*
*
******************************************************/
bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
                     vector<int8> &data)
{
     

  
    // Get the size of the table
    vector<unsigned long> size(1);
    size[0] = 1;
    size[1] = nData;
  
   
    // Create the column
    Column *column = table->addColumn( columnName, Dal::Int8, comment, units , size);
 
    ColumnData *columnData = column->data();

    setColumnData( ( int)size[1], columnData, data);


    delete columnData;


          
    return true;
};


/********************************************************
*
*
******************************************************/
bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
                     vector<int16> &data)
{
     

  
    // Get the size of the table
    vector<unsigned long> size(1);
    size[0] = 1;
    size[1] = nData;
  
   
    // Create the column
    Column *column = table->addColumn( columnName, Dal::Int16, comment, units , size);
 
    ColumnData *columnData = column->data();

    setColumnData( ( int)size[1], columnData, data);


    delete columnData;


          
    return true;
};



/********************************************************
*
*
*
*
*
******************************************************/
bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
                     vector<int32> &data)
{
     

  
    // Get the size of the table
    vector<unsigned long> size(1);
    size[0] = 1;
    size[1] = nData;
  
   
    // Create the column
    Column *column = table->addColumn( columnName, Dal::Int32, comment, units , size);
 
    ColumnData *columnData = column->data();

    setColumnData( ( int)size[1], columnData, data);

    delete columnData;
          
    return true;
};


/******************************************************
*
*
*
******************************************************/
void setColumnData( const int size, const ColumnData *columnData, const vector<int8>& data)
{
    int8 * r64 = columnData->int8Data();
    for(int i = 0; i < size; i++)
        r64[ i] = data[i];


}




/******************************************************
*
*
*
******************************************************/
void setColumnData( const int size, const ColumnData *columnData, const vector<int16>& data)
{
    int16 * r64 = columnData->int16Data();
    for(int i = 0; i < size; i++)
        r64[ i] = data[i];


}


/******************************************************
*
*
*
******************************************************/
void setColumnData( const int size, const ColumnData *columnData, const vector<int32>& data)
{
    int32 * r64 = columnData->int32Data();
    for(int i = 0; i < size; i++)
        r64[ i] = data[i];


}








/******************************************************
*
******************************************************/


void setColumnData( const int size, const ColumnData *columnData, const vector<real32>& data )
{
    real32 * r32 = columnData->real32Data();
    for(int i = 0; i < size; i++)
        r32[i] = data[i];

}

/******************************************************
*
******************************************************/
void setColumnData( const int size, const ColumnData *columnData, const vector<real64>& data)
{
    real64 * r64 = columnData->real64Data();
    for(int i = 0; i < size; i++)
        r64[ i] = data[i];


}





 
/********************************************************
*
*
*******************************************************/
void addSkyattributes(string file, int x0, int y0, float ra, float dec)
{

    DataSet *dataSet = dataSetServer->open( file, Dal::Modify);
 

    *dataSet->addAttribute("RA_SCX", "") = ra;

    *dataSet->addAttribute("RA_PNT", "") = ra;

    *dataSet->addAttribute("DEC_SCX", "") = dec;

    *dataSet->addAttribute("DEC_PNT", "") = dec;
    *dataSet->addAttribute("PA", "") = 45.0;
    *dataSet->addAttribute("EQUINOX", "") = 2000.0;

    *dataSet->addAttribute("CTYPE1", "") =  "RA---TAN";
    *dataSet->addAttribute("CTYPE2", "") =  "DEC---TAN";

    *dataSet->addAttribute("CDELT1", "") =  -0.0001;
    *dataSet->addAttribute("CDELT2", "") =  0.0001;

    *dataSet->addAttribute("CRPIX1", "") =  x0;
    *dataSet->addAttribute("CRPIX2", "") =  y0;


    *dataSet->addAttribute("CRVAL1", "") =  ra;
    *dataSet->addAttribute("CRVAL2", "") =  dec;
 
   dataSetServer->close( dataSet );

}      
